/**
 * PROJEKT: 	Java native access to K+K Library (kk_fx80e, kk_library_64) Windows and Linux
 * UNIT:	    KK_Library.java
 * COPYRIGHT:	Brendes Datentechnik 2017..2021, Loryn Brendes
 * 				www.brendes.de
 * 
 *			-- ABSTRACT --
 *
 * Defines interface to K+K library minimal version 19.3.1
 * 	kk_library_64 on 64-bit systems
 *	kk_fx80e	  on 32-bit systems	
 *
 * 
 *          -- HINT --
 *
 * JAR files jna-4.2.2.jar and jna-platform-4.2.2.jar are needed
 * Please adjust path entries in file .classpath.
 * 
 * Predefined path="C:/Program Files/Java/"
 *  
 *  
 * 			-- HISTORY --
 *
 * DD.MM.YY who	what
 * 28.09.16	lb	created DLL version 16.03
 * 22.11.16	lb	added: KK_GetHostAndIPs
 * 17.12.16	lb	DLL version 16.04: KK_Debug replaces FX_Debug, added: KK_SetOutputPath
 * 21.12.16	lb	Version 1.0 for KK library kk_fx80e version 16.04
 * 06.02.17	lb	Version 1.1 for KK library kk_fx80e version 16.05, throws IllegalAccessException
 * 21.11.17	lb	Version 1.2 for KK library kk_fx80e version 16.10:
 * 					bug fixes bufferToString(), sendCommand(), buffer-0-byte
 * 					openPort with blockingIO, contextMode
 * 					getToolsData()
 * 15.11.18 lb  Version 1.3 for KK library kk_fx80e version 18.0:
 * 					only use Multi_* function calls
 * 12.10.20 lb	Version 1.4 for KK library kk_fx80e version 18.1.10:
 * 					added: getFirmwareVersion, hasFRAM, setNSZCalibrationData, remoteLogin
 *                  TCP-Server NSZ modes, logtypes added
 * 					select proper library name
 * 28.05.21 lb	Version 1.5 for K+K library min version 19.0.2
 * 					added TCP-Server modes, logtypes: PhasePredecessor, User1/2
 * 					added: Error Reconnected (getReport)
 * 					added: openTcpLogTime
 * 18.11.21 lb	Version 1.6 for K+K library min version 19.1.2
 * 					added: ENUM_FLAG_PCAN
 * 					added: helper classes FHR_Data, FHR_Settings
 * 					added: getFHRSettings, setFHRSettings
 * 					added: Linux support
 * 20.12.21 lb  Version 1.7 for K+K library min version 19.2.0
 *                  added: openTcpLogType, tcpAppData
 * 09.10.22 lb  Version 1.8 for K+K library min version 19.3.0
 * 05.01.23 lb  Version 1.9 for K+K library min version 19.3.1
 */

package brendes.jna;

import java.nio.charset.StandardCharsets;
import java.util.Formatter;
import java.util.Locale;
import java.util.regex.Pattern;

import com.sun.jna.Library;
import com.sun.jna.Native;

/**
 * KK_Library class gives native access to K+K library (kk_fx80e for 32-bit systems, 
 * kk_library_64 for 64-bit systems) in Windows and Linux.
 * K+K library must have version 19.3.1 or higher.
 * 
 * @author Loryn
 *
 */
public final class KK_Library  {

	/*
	 ***************************************************************************
	 * Constructor
	 ***************************************************************************
	 */
	
	public String dllVersion;
	public String dllDate;
	
	/**
	 * Creates instance of class KK_Library. Checks version number of K+K library.
	 * @throws IllegalAccessException if found version is less than 19.3.1
	 */
	public KK_Library() throws IllegalAccessException {
		super();
		
		// check version of K+K library
		String dllVersionDate = getDLLVersion();
		// split into version date
		String[] sub = dllVersionDate.split(" ");
		if (sub.length < 2)
			throw new UnsupportedClassVersionError(dllVersionDate);
		dllVersion = sub[0];
		dllDate = sub[1];
		// split version numbers 
		// "." is regular expression -> escape it
		String s = Pattern.quote(".");
		sub = dllVersion.split(s);
		int major = 0;
		int minor = 0;
		int release = 0;
		try {
			major = Integer.parseInt(sub[0]);
			if (sub.length > 1)
				minor = Integer.parseInt(sub[1]);
			if (sub.length > 2)
				release = Integer.parseInt(sub[2]);
		}
		catch (Exception e) {
			// ignore
		}
		boolean versionOk = true;	// 19.3.1
		if (major < 19)
			versionOk = false;
		else if (major == 19) {
			if (minor < 2)
				versionOk = false;
			else if (minor == 3) {
				if (release < 1)
					versionOk = false;
			}
		}
		if (!versionOk)
			throw new IllegalAccessException("Access to library "+LIB_NAME
					+" needs version 19.3.1 or higher. Version found: "+dllVersion);
		
		// resources
		fhrData = new FHR_Settings();
}
	
	/*
	 ***************************************************************************
	 * Version
	 ***************************************************************************
	 */
	
	/**
	 * Version number and date of class KK_Library
	 */
	public static final String VERSION = "1.8 2022-10-09";
	
	/*
	 ***************************************************************************
	 * Library interface
	 ***************************************************************************
	 */

	public static String LIB_NAME = getLibName();
	
	public static String getLibName() {
		String s = System.getProperty("os.arch");	// Architecture of Java-VM
		boolean is64 = s.indexOf("64") > -1;
		s = System.getProperty("os.name").toLowerCase(Locale.ENGLISH);
		boolean isWindows = s.indexOf("win") > -1;

		if (isWindows) {
			if (is64)
				return "kk_library_64";
			else
				return  "kk_fx80e";
		}
		else {
			if (is64)
				return "libkk_library_64.so";
			else
				return  "libkk_fx80e.so";
		}
	}

	/**
	 * Interface to native K+K library kk_fx80e, kk_library_64 in Windows and Linux.
	 * Defines exported functions of native K+K library.
	 * <br><b>Hint</b> Java <code>String</code> is for NULL terminated const char* only!
	 * @author Loryn
	 *
	 */
	interface KK_Lib extends Library {
		
		final static KK_Lib INSTANCE = (KK_Lib)Native.loadLibrary(LIB_NAME, KK_Lib.class);

		/*
		 * create multi source object
		 */
		int CreateMultiSource();
		
		/*
		 * enumerate local devices (serial, USB)
		 */
		int Multi_EnumerateDevices(byte[] Names, byte EnumFlags);
		String Multi_GetEnumerateDevicesErrorMsg();

		/*
		 * enumerate local IPs
		 */
		int Multi_GetHostAndIPs(byte[] HostName, byte[] IPaddr, byte[] ErrorMsg);
		
		/*
		 * output path
		 */
		String Multi_GetOutputPath(int ID);
		String Multi_SetOutputPath(int ID, String Path);
		
		/*
		 * Debug
		 */
		String Multi_Debug(int ID, Boolean dbgOn, String DbgID);
		int Multi_DebugFlags(int ID, Boolean ReportLog, Boolean LowLevelLog);
		int Multi_DebugLogLimit(int ID, byte LogType, int aSize);
		String Multi_DebugGetFilename(int ID);

		/*
		 * information requests
		 */
		String Multi_GetDLLVersion();
		int Multi_GetBufferAmount(int ID);
		int Multi_GetTransmitBufferAmount(int ID);
		byte Multi_GetUserID(int ID);
		Boolean Multi_IsFileDevice(int ID);
		Boolean Multi_IsSerialDevice(int ID);
		int Multi_GetFirmwareVersion(int ID);
		Boolean Multi_HasFRam(int ID);
		int Multi_GetDeviceStartState(int ID);

		/*
		 * NSZ Calibration
		 */
		int Multi_SetNSZCalibrationData(int ID, String Data);
		
		/*
		 * FHR Settings
		 */
		int Multi_ReadFHRData(int ID);
		int Multi_SetFHRData(int ID, String Data);
		
		/*
		 * open, close
		 */
		int Multi_OpenConnectionEx(int ID, byte[] Connection, Boolean blockingIO, Boolean contextMode,
				Object aObject);
		void Multi_CloseConnection(int ID);
		
		/*
		 * read, write
		 */
		int Multi_SetDecimalSeparator(int ID, byte Separator);
		int Multi_SetNSZ(int ID, int aNSZ);
		int Multi_GetReport(int ID, byte[] Data);
		int Multi_SetSend7016(int ID, boolean value);
		int Multi_GetToolsData(int ID, byte[] Data);
		int Multi_GetPendingCmdsCount(int ID);
		int Multi_SetCommandLimit(int ID, int Limit);
		int Multi_SendCommand(int ID, byte[] Command, int len);
		int Multi_RemoteLogin(int ID, int Password, byte[] err);
		
		/*
		 * TCP server
		 */
		int Multi_StartTcpServer(int ID, short[] aPort);
		int Multi_StopTcpServer(int ID);
		String Multi_GetTcpServerError(int ID);
		void Multi_TcpReportLog(int ID, String Data, int logType);
		
		/* 
		 * connection to TCP server
		 */
		int Multi_OpenTcpLog(int ID, byte[] IpPort, String Mode);
		int Multi_OpenTcpLogTime(int ID, byte[] IpPort, String Mode, String formatTime);
		int Multi_OpenTcpLogType(int ID, byte[] IpPort, int LogType, String formatTime);
		void Multi_CloseTcpLog(int ID);
		int Multi_GetTcpLog(int ID, byte[] Data);
		int Multi_TcpAppData(int ID, byte[] Data);

		/*
		 * generate test data
		 */
		// Binary files
		int Multi_StartSaveBinaryData(int ID, String DbgID);
		int Multi_StopSaveBinaryData(int ID);
		// Text files
		int Multi_StartSaveReportData(int ID, String DbgID);
		int Multi_StopSaveReportData(int ID);

	}

	/*
	 ***************************************************************************
	 * Library return class
	 ***************************************************************************
	 */

	/**
	 * Defines result code values for class {@link KK_Result}
	 * @author Loryn
	 *
	 */
	public enum ErrorCode {
		/**
		 * successful operation
		 */
		KK_No_Err,
		/**
		 * operation failed
		 * {@link KK_Result#data} contains error message
		 */
		KK_Err,
		/**
		 * Enumeration of serial ports failed
		 * {@link KK_Result#data} contains error message
		 */
		KK_Err_Enum_Serial,
		/**
		 * Enumeration of USB devices failed
		 * {@link KK_Result#data} contains error message
		 */
		KK_Err_Enum_USB,
		/**
		 * Enumeration of serial ports and USB devices failed
		 * {@link KK_Result#data} contains error message
		 */
		KK_Err_Enum_Serial_USB,
		/**
		 * Enumeration of PCAN-USB devices failed
		 * {@link KK_Result#data} contains error message
		 */
		KK_Err_Enum_PCAN,
		/**
		 * String in {@link KK_Result#data} truncated to 1024 character
		 */
		KK_Err_BufferTooSmall,
		/**
		 * Lost of data occurred. App does not read fast enough
		 */
		KK_Err_BufferOverflow,
		/**
		 * Writing via current connection failed
		 * {@link KK_Result#data} contains error message
		 */
		KK_Err_Write,
		/**
		 * Connection to K+K server has broken, server is down
		 * {@link KK_Result#data} contains error message
		 */
		KK_Err_ServerDown,
		/**
		 * Connection to K+K device is interrupted (no usb cable)
		 */
		KK_Err_DeviceNotConnected,
		/**
		 * K+K device does not send measurement data. Measurement hardware should be checked.
		 * Connection to K+K device is closed. {@link KK_Result#data} contains error message
		 */
		KK_HardwareFault,
		/**
		 * Parameter in function call has invalid value
		 */
		KK_ParamError,
		/**
		 * Command rejected (no connection, waiting queue full)
		 */
		KK_CmdIgnored,
		/**
		 * Requested function is not supported by K+K device
		 */
		KK_NotSupported,
		/**
		 * Reconnection of an interrupted connection with data loss
		 * @since 19.00.02
		 */
		KK_Reconnected
	}

	/**
	 * Return type of most KK_Library methods
	 * @author Loryn
	 *
	 */
	public class KK_Result {
		/**
		 * Operation result code see {@link ErrorCode}
		 */
		public ErrorCode resultCode;
		/**
		 * Operation string result or error message if {@link #resultCode} != <code>KK_No_Err</code>
		 */
		public String data;
		/**
		 * Contents of data split into several lines
		 */
		public String[] lines;
		/**
		 * An integer return value, set by some methods (z.B. {@link KK_Library#startTcpServer(int, short)})
		 */
		public int intValue;
	}
	
	/*
	 * Error strings from K+K library
	 */
	private static final String ERR_SOURCE_NOT_FOUND = "Source-ID ";
	
	/*
	 * Error numbers from K+K library
	 */
	private static final int CKK_DLL_Error               = 0;
	private static final int CKK_DLL_NoError             = 1;
	private static final int CKK_DLL_WriteError          = 3;
	private static final int CKK_DLL_ServerDownError     = 4;
	/**
	 * Response string truncated to CKK_DLL_BufferSize characters
	 */
	private static final int CKK_DLL_BufferTooSmall      = 6;  
	/**
	 * CDC: USB cable was unplugged
	 */
	private static final int CKK_DLL_DeviceNotConnected  = 7;
	/**
	 * Overflow receive buffer -> data loss
	 */
	private static final int CKK_DLL_BufferOverflow      = 8;
	/**
	 * K+K device doesn't send measurement data, connection closed
	 */
	private static final int CKK_DLL_HardwareFault       = 9;
	/**
	 * MultiSource: invalid Source-ID
	 */
	private static final int CKK_DLL_SourceNotFound      = 10;
	/**
	 * Command rejected (no connection, command list full)
	 */
	private static final int CKK_DLL_CmdIgnored          = 11;
	/**
	 * K+K device does not support the requested function
	 */
	private static final int CKK_DLL_NotSupported		 = 12;
	/**
	 * Reconnection of an interrupted connection with data loss
	 * @since 19.00.02
	 */
	private static final int CKK_DLL_Reconnected         = 13;

	/*
	 ***************************************************************************
	 * buffer
	 ***************************************************************************
	 */

	/**
	 * buffer for data transmission between KK_Library class and K+K library
	 */
	private byte[] buffer = new byte[1024];
	
	/**
	 * Converts aBuffer contents into String
	 * @param aBuffer with bytes to convert
	 * @return aBuffer contents as String
	 */
	private String bufferToString(byte[] aBuffer) {
		String result = null;
		// buffer contents must be ASCII, stop at first 0 byte
		int index = 0;
		for (int i = 0; i < aBuffer.length; i++)
			if (aBuffer[i] == 0) {
				index = i;
				break;
			}
		if (index > 0) {
			// convert back to string
			byte[] sBytes = new byte[index];
			System.arraycopy(aBuffer, 0, sBytes, 0, index);
			result = new String(sBytes, StandardCharsets.US_ASCII);
		}
		return result;
	}

	
	/*
	 ***************************************************************************
	 * enumerate local devices (serial, USB)
	 ***************************************************************************
	 */

	/**
	 * Flag for {@link #enumerateDevices(byte)}: enumerate serial ports only
	 */
	public final static byte ENUM_FLAG_SERIAL_PORTS	= 0x01;
	/**
	 * Flag for {@link #enumerateDevices(byte)}: enumerate K+K devices found on USB
	 */
	public final static byte ENUM_FLAG_USB			= 0x02;
	/**
	 * Flag for {@link #enumerateDevices(byte)}: enumerate serial ports and K+K devices on USB
	 */
	public final static byte ENUM_FLAG_LOCAL_DEVICES	= 0x03;
	/**
	 * Flag for {@link #enumerateDevices(byte)}: enumerate CAN devices via PCAN-USB-Adapter 
	 * @since 18.01.03
	 */
	public final static byte ENUM_FLAG_PCAN			= 0x04;
	

	
	/**
	 * Enumerates K+K devices that are locally connected to the PC (serial, USB, PCAN-USB), 
	 * depending on <code>enumFlags</code>.
	 * Found serial port names, USB devices names and PCAN-USB are returned in KK_Result.data 
	 * separated by comma (,) and split into lines in KK_Result.lines.
	 * When no names are found KK_Result.data and KK_Result.lines are <code>null</code>.
	 * If an error occurs, KK_Result.resultCode is one of KK_Err_Enum_*, KK_Result.data contains 
	 * error message.
	 * <br>For details see <b>Multi_EnumerateDevices</b> in K+K library manual.
	 * 
	 * @param enumFlags	specifies what is to be enumerated: {@link KK_Library#ENUM_FLAG_SERIAL_PORTS},
	 * {@link KK_Library#ENUM_FLAG_USB}, {@link KK_Library#ENUM_FLAG_LOCAL_DEVICES}, 
	 * {@link KK_Library#ENUM_FLAG_PCAN}
	 * 	
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> KK_Result.data contains found names separated by comma (,) or 
	 * <code>null</code>, if no serial ports and/or USB devices were found. KK_Result.lines contains
	 * found single names, if at least one name is found. Otherwise KK_Result.lines is <code>null</code>.
	 * <br><b>KK_ParamError:</b> Parameter enumFlags has invalid value. 
	 * KK_Result.data contains error message.
	 * <br><b>KK_Err_Enum_Serial, KK_Err_Enum_USB, KK_Err_Enum_Serial_USB, KK_Err_Enum_PCAN:</b> enumeration failed,
	 * error occurred on enumeration of serial ports and/or USB devices.
	 * KK_Result.data contains error message.
	 * 
	 * @since K+K library version 18.0 
	 */
	public KK_Result enumerateDevices(byte enumFlags) {
		KK_Result result = new KK_Result();
		
		switch (enumFlags) {
			case ENUM_FLAG_LOCAL_DEVICES:
			case ENUM_FLAG_PCAN:
			case ENUM_FLAG_SERIAL_PORTS:
			case ENUM_FLAG_USB: break;	// ok
			default: {
			  result.resultCode = ErrorCode.KK_ParamError;
			  result.data = "Parameter enumFlags has invalid value";
			  return result;
			}
		}
		
		int res = KK_Lib.INSTANCE.Multi_EnumerateDevices(buffer, enumFlags);
		if (res == 0) {
			// enumeration successful
			result.resultCode = ErrorCode.KK_No_Err;
			result.data = bufferToString(buffer);
			if (result.data != null)
				result.lines = result.data.split(",");
		} else {
			if ((res & ENUM_FLAG_SERIAL_PORTS & ENUM_FLAG_USB) != 0)
				result.resultCode = ErrorCode.KK_Err_Enum_Serial_USB;
			else if ((res & ENUM_FLAG_SERIAL_PORTS) != 0)
				result.resultCode = ErrorCode.KK_Err_Enum_Serial;
			else if ((res & ENUM_FLAG_USB) != 0)
				result.resultCode = ErrorCode.KK_Err_Enum_USB;
			else if ((res & ENUM_FLAG_PCAN) != 0)
				result.resultCode = ErrorCode.KK_Err_Enum_PCAN;
			// deliver error message in data
			result.data = KK_Lib.INSTANCE.Multi_GetEnumerateDevicesErrorMsg();
		}
		
		return result;
	}
	
	/*
	 ***************************************************************************
	 * enumerate local IPs
	 ***************************************************************************
	 */

	/**
	 * Enumerates local IPv4 addresses and host name of local computer and delivers
	 * them in <code>hostName</code> and <code>ipAddr</code>. If an error occurs KK_Result.data
	 * contains error message and <code>hostName[0]</code> and/or <code>ipAddr[0]</code>
	 * contains 0.
	 * <br>For details see <b>Multi_GetHostAndIPs</b> in K+K library manual.
	 * 
	 * @param hostName buffer must have minimum 80 bytes, contains host name of local computer 
	 * after call
	 * @param ipAddr buffer must have minimum 80 bytes, contains local IPv4 addresses separated by
	 * comma after call
	 * 
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> successful operation 
	 * <br><b>KK_Err_BufferTooSmall:</b> successful operation, but string in <code>hostName</code> 
	 * and/or <code>ipAddr</code> is truncated to 80 character
	 * <br><b>KK_Err</b> enumeration failed, KK_Result.data contains error message
	 * 
	 * @since K+K library version 18.0
	 */
	public KK_Result enumerateLocalIPs(byte[] hostName, byte[] ipAddr) {
		KK_Result result = new KK_Result();

		// need 80 bytes per byte array
		if ((hostName == null) || (hostName.length < 80)) {
			result.resultCode = ErrorCode.KK_ParamError;
			result.data = "Parameter hostName has invalid value";
			return result;
		}
		if ((ipAddr == null) || (ipAddr.length < 80)) {
			result.resultCode = ErrorCode.KK_ParamError;
			result.data = "Parameter ipAddr has invalid value";
			return result;
		}
		
		int res = KK_Lib.INSTANCE.Multi_GetHostAndIPs(hostName, ipAddr, buffer);
		switch (res) {
		case CKK_DLL_NoError:
			result.resultCode = ErrorCode.KK_No_Err;
			break;
		case CKK_DLL_BufferTooSmall:
			result.resultCode = ErrorCode.KK_Err_BufferTooSmall;
			break;
		default:
			result.resultCode = ErrorCode.KK_Err;
			result.data = bufferToString(buffer);
		}
				
		return result;
	}
	
	/**
	 * Convenience call
	 * Enumerates local IPv4 addresses and host name of local computer and delivers
	 * them in KK_Result.data (host name) and KK_Result.lines (IPv4 addresses). 
	 * If an error occurs KK_Result.data contains error message.
	 * <br>For details see <b>Multi_GetHostAndIPs</b> in K+K library manual.
	 * 
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> successful operation, KK_Result.data contains host name and KK_Result.lines
	 * contains IPv4 addresses. 
	 * <br><b>KK_Err_BufferTooSmall:</b> successful operation, but string in KK_Result.data 
	 * and/or KK_Result.lines is truncated to 80 character
	 * <br><b>KK_Err</b> enumeration failed, KK_Result.data contains error message
	 * 
	 * @since K+K library version 18.0
	 */
	public KK_Result enumerateLocalIPs() {
		KK_Result result = new KK_Result();
		byte[] hostName = new byte[80];
		byte[] ipAddr = new byte[80];

		int res = KK_Lib.INSTANCE.Multi_GetHostAndIPs(hostName, ipAddr, buffer);
		switch (res) {
		case CKK_DLL_NoError:
			result.resultCode = ErrorCode.KK_No_Err;
			result.data = bufferToString(hostName);
			String s = bufferToString(ipAddr);
			if (s != null)
				result.lines = s.split(",");
			break;
		case CKK_DLL_BufferTooSmall:
			result.resultCode = ErrorCode.KK_Err_BufferTooSmall;
			break;
		default:
			result.resultCode = ErrorCode.KK_Err;
			result.data = bufferToString(buffer);
		}
				
		return result;
	}
	
	/*
	 ***************************************************************************
	 * create multi source
	 ***************************************************************************
	 */
	
	/**
	 * Creates new multi source management object (source) and returns identifier 
	 * of this object. This identifier must be used with subsequent library calls.
	 * <br><b>Hint:</b> All generated management object are released on program end.
	 * <br>For details see <b>CreateMultiSource</b> in K+K library manual.
	 * 
	 * @return identifier of library internal management object
	 * 
	 * @since K+K library version 18.0
	 */
	public int createMultiSource() {
		return KK_Lib.INSTANCE.CreateMultiSource();
	}
	
	/*
	 ***************************************************************************
	 * output path
	 ***************************************************************************
	 */

	/**
	 * Get current output path for files generated by K+K library (test data, debug log)
	 * for source <code>id</code>.
	 * <br>For details see <b>Multi_GetOutputPath</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()}
	 *  
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> KK_Result.data contains current output path for source id 
	 * <br><b>KK_ParamError:</b> invalid id, KK_Result.data contains error message.
	 * 
	 * @since K+K library version 18.0
	 */
	public KK_Result getOutputPath(int id) {
		KK_Result result = new KK_Result();

		result.data = KK_Lib.INSTANCE.Multi_GetOutputPath(id);
		if (result.data == null)
			result.resultCode = ErrorCode.KK_No_Err;
		else if (result.data.startsWith(ERR_SOURCE_NOT_FOUND))
			result.resultCode = ErrorCode.KK_ParamError;

		return result;
	}
	
	/**
	 * Set output path for files generated by K+K library (test data, debug log)
	 * for source <code>id</code>. New output path is only valid for files generated in 
	 * future. Still generated files keep their path names.
	 * <br>For details see <b>Multi_SetOutputPath</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * @param path pathname of output directory
	 * 
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> successful operation 
	 * <br><b>KK_ParamError:</b> invalid id, KK_Result.data contains error message
	 * <br><b>KK_Err:</b> call failed, KK_Result.data contains error message
	 * 
	 * @since K+K library version 18.0
	 */
	public KK_Result setOutputPath(int id, String path) {
		KK_Result result = new KK_Result();

		result.data = KK_Lib.INSTANCE.Multi_SetOutputPath(id, path);
		if (result.data == null)
			result.resultCode = ErrorCode.KK_No_Err;
		else if (result.data.startsWith(ERR_SOURCE_NOT_FOUND))
			result.resultCode = ErrorCode.KK_ParamError;
		else 
			result.resultCode = ErrorCode.KK_Err;
		
		return result;
	}

	/*
	 ***************************************************************************
	 * Debug
	 ***************************************************************************
	 */

	/**
	 * Switches on/off debug log outputs of K+K library for source <code>id</code>.
	 * If <code>dbgOn</code> is true K+K library generates a debug log in file
	 * <b>KK_DLL_LOG_<code>DbgID</code>.txt</b>. The file is located in path set by 
	 * {@link #setOutputPath(int, String)} or in <i>application path</i>. 
	 * A still existing file will be overwritten.
	 * Please see {@link #setDebugLogLimit(int, DebugLogType, int)} for additional debug
	 * settings.
	 * Every call to a library function generates log outputs to the debug file.
	 * If <code>dbgOn</code> is false debug log file is closed.
	 * <br><b>Hint:</b> <code>DbgID</code> is used to get unique file names for
	 * distinct sources. This is needed in particular, if common output paths are used.
	 * <code>dbgID</code> maybe null, if only one source is used.
	 * <br>For details see <b>Multi_Debug</b> in K+K library manual.
	 *  
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * @param dbgOn <code>true</code> opens log file, <code>false</code> close log file
	 * @param dbgID source identifier of debug log file, part of file name
	 * 
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> successful operation 
	 * <br><b>KK_ParamError:</b> invalid id, KK_Result.data contains error message
	 * <br><b>KK_Err:</b> call failed, KK_Result.data contains error message
	 * 
	 * @since K+K library version 18.0
	 */
	public KK_Result setDebug(int id, boolean dbgOn, String dbgID) {
		KK_Result result = new KK_Result();

		result.data = KK_Lib.INSTANCE.Multi_Debug(id, dbgOn, dbgID);
		if (result.data == null)
			result.resultCode = ErrorCode.KK_No_Err;
		else if (result.data.startsWith(ERR_SOURCE_NOT_FOUND))
			result.resultCode = ErrorCode.KK_ParamError;
		else 
			result.resultCode = ErrorCode.KK_Err;
		return result;
	}
	
	/**
	 * Sets amount of data which is logged for source <code>id</code>. 
	 * Changed values take effect immediately.
	 * <br>For details see <b>Multi_DebugFlags</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * @param reportLog generate log output about received reports
	 * @param lowLevelLog generate log output about byte stream received from/send to K+K device
	 * or K+K server.
	 * 
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> successful operation 
	 * <br><b>KK_ParamError:</b> invalid id
	 * 
	 * @since K+K library version 18.0
	 */
	public KK_Result setDebugFlags(int id, boolean reportLog, boolean lowLevelLog) {
		KK_Result result = new KK_Result();
		result.resultCode = ErrorCode.KK_No_Err;

		int res = KK_Lib.INSTANCE.Multi_DebugFlags(id, reportLog, lowLevelLog);
		if (res == CKK_DLL_SourceNotFound) {
			result.resultCode = ErrorCode.KK_ParamError;
		}
		
		return result;
	}
	
	/**
	 * Defines debug log types for method {@link KK_Library#setDebugLogLimit(int, DebugLogType, int)}
	 * Specifies file management of debug output files.
	 * @author Loryn
	 *
	 */
	public enum DebugLogType {
		/**
		 * All debug outputs are written to the same file, reopen overwrites existing debug file.
		 * File size is unlimited. Default value
		 */
		logUnlimited,
		/**
		 * All debug outputs are written to the same file, reopen overwrites existing debug file.
		 * File size is limited. If size is reached old debug outputs will be overwritten.
		 */
		logOverwrite,
		/**
		 * A new debug file is created, filename is appended by date and time stamp.
		 * File size is limited. If size is reached a new file is created.
		 */
		logCreateNew
	}
	
	/**
	 * Sets limits of debug log file for source <code>id</code>. 
	 * If aSize = 0 default size of 20 MB is used.
	 * Default DebugLogType is logUnlimited.
	 * <br>For details see <b>Multi_DebugLogLimit</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * @param logType specifies debug file management, a value of {@link DebugLogType}
	 * @param aSize maximum size of debug log file, when logType != logUnlimited
	 * 
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> successful operation 
	 * <br><b>KK_ParamError:</b> invalid id
	 * 
	 * @since K+K library version 18.0
	 */
	public KK_Result setDebugLogLimit(int id, DebugLogType logType, int aSize) {
		KK_Result result = new KK_Result();
		result.resultCode = ErrorCode.KK_No_Err;
		byte type;
		if (logType == DebugLogType.logOverwrite)
			type = 1;
		else if (logType == DebugLogType.logCreateNew)
			type = 2;
		else type = 0;
		
		int res = KK_Lib.INSTANCE.Multi_DebugLogLimit(id, type, aSize);
		if (res == CKK_DLL_SourceNotFound) {
			result.resultCode = ErrorCode.KK_ParamError;
		}
		
		return result;
	}

	/**
	 * Gets filename of debug log file for source <code>id</code>.
	 * <br>For details see <b>Multi_DebugGetFilename</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * 
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> KK_Result.data contains filename or KK_Result.data = null in case
	 * debug log is switched off
	 * <br><b>KK_ParamError:</b> invalid id, KK_Result.data contains error message
	 * 
	 * @since K+K library version 18.0
	 */
	public KK_Result getDebugFilename(int id) {
		KK_Result result = new KK_Result();
		result.resultCode = ErrorCode.KK_No_Err;
		
		result.data = KK_Lib.INSTANCE.Multi_DebugGetFilename(id);
		if (result.data != null && result.data.startsWith(ERR_SOURCE_NOT_FOUND))
			result.resultCode = ErrorCode.KK_ParamError;
		
		return result;
	}
	
	/*
	 ***************************************************************************
	 * information requests
	 ***************************************************************************
	 */

	/**
	 * Returns string with native K+K libraries version and date.
	 * <br>For details see <b>Multi_GetDLLVersion</b> in K+K library manual.
	 * 
	 * @return String with library version and date
	 * 
	 * @since K+K library version 18.0
	 */
	public String getDLLVersion() {
		return  KK_Lib.INSTANCE.Multi_GetDLLVersion();
	}
	
	/**
	 * Gets receive buffer fill level for source <code>id</code> in count of bytes.
	 * <br>For details see <b>Multi_GetBufferAmount</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * 
	 * @return 
	 * 0: receive buffer is empty or error (invalid id, no current connection)
	 * <br>0: count of bytes in receive buffer for source <code>id</code>
	 *   
	 * @since K+K library version 18.0
	 */
	public int getBufferAmount(int id) {
		return KK_Lib.INSTANCE.Multi_GetBufferAmount(id);
	}

	/**
	 * Gets transmit buffer fill level of K+K device connected to source <code>id</code>.
	 * K+K device can manage up to four users. Every user has its own transmit buffer. 
	 * The transmit buffer fill level is transmitted by connections via USB and to remote devices.
	 * <br>For details see <b>Multi_GetTransmitBufferAmount</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * 
	 * @return 
	 * <br>0: transmit buffer is empty or error (invalid id, no connection to K+K device)
	 * <br>0: count of bytes in transmit buffer still not send for source <code>id</code>
	 *   
	 * @since K+K library version 18.0
	 */
	public int getTransmitBufferAmount(int id) {
		return KK_Lib.INSTANCE.Multi_GetTransmitBufferAmount(id);
	}

	/**
	 * Returns user id assigned by K+K device of current connection for source <code>id</code>.
	 * A K+K device is able to serve up to 4 users simultaneously. The users are numbered from 1 
	 * to 4. The user id is transmitted by connections via USB and via network.
	 * Newer firmware versions of K+K devices transmit user id within the version string
	 * (reaction of command 0x01 or 0x81).
	 * <br>For details see <b>Multi_GetUserID</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()}
	 *  
	 * @return 
	 * <br>1..4: user id assigned by K+K device
	 * <br>0xFF: error (invalid id, no connection to K+K device) 
	 *   
	 * @since K+K library version 18.0
	 */
	public byte getUserID(int id) {
		return KK_Lib.INSTANCE.Multi_GetUserID(id);
	}
	
	/**
	 * If data is read from file for source <code>id</code>, this function returns true.
	 * <br>For details see <b>Multi_IsFileDevice</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * 
	 * @return data returned by {@link #getReport(int)} is read from file
	 *   
	 * @since K+K library version 18.0
	 */
	public Boolean isFileDevice(int id) {
		return KK_Lib.INSTANCE.Multi_IsFileDevice(id);
	}
	
	/**
	 * If data is read from serial connection for source <code>id</code>, this function returns true.
	 * <br>For details see <b>Multi_IsSerialDevice</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * 
	 * @return data returned by {@link #getReport(int)} is read from serial connection
	 *   
	 * @since K+K library version 19.1.2
	 */
	public Boolean isSerialDevice(int id) {
		return KK_Lib.INSTANCE.Multi_IsSerialDevice(id);
	}
	
	/**
	 * Returns firmware version number of last version report received from K+K device 
	 * connected to source <code>id</code>.
	 * To request a version report, send command 0x01 or 0x81 to K+K device.
	 * <br>For details see <b>Multi_GetFirmwareVersion</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()}
	 * 
	 * @return
	 * <br>-1: error (invalid id, no version report received)
	 * <br> -1: firmware version number
	 *   
	 * @since K+K library version 18.1.10
	 */
	public int getFirmwareVersion(int id) {
		return KK_Lib.INSTANCE.Multi_GetFirmwareVersion(id);
	}
	
	/**
	 * Returns K+K device connected to source <code>id</code> is equipped with F-RAM.
	 * This feature is reported by flag F in version report. Therefore a version
	 * report must have been received.
	 * To request a version report, send command 0x01 or 0x81 to K+K device.
	 * <br>For details see <b>Multi_HasFRAM</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()}
	 * 
	 * @return
	 * <br>false: K+K device has no F-RAM, or no version report received, or invalid id
	 * <br>true: K+K device uses F-RAM to store NSZ calibration data
	 *   
	 * @since K+K library version 18.1.10
	 */
	public Boolean hasFRAM(int id) {
		return KK_Lib.INSTANCE.Multi_HasFRam(id);
	}
	
	/**
	 * Delivers the state of the K+K device at application start. 
	 * The start state is determined from the first received timestamp.
	 * <br>For details see <b>Multi_GetDeviceStartState</b> in K+K library manual.
	 * @param id source identifier returned by {@link #createMultiSource()}
	 * 
	 * @return
	 * <br>0: Source-ID invalid or no time stamp received
	 * <br>1: Cold start, K+K device was restarted
	 * <br>2: Warm start
	 * 
	 * @since K+K library version 19.3.1
	 */
	public int getDeviceStartState(int id) {
		return KK_Lib.INSTANCE.Multi_GetDeviceStartState(id);
	}

	
	/*
	 ***************************************************************************
	 * NSZ Calibration
	 ***************************************************************************
	 */

	/**
	 * Sets NSZ calibration values <code>calibrationValues</code> in K+K device 
	 * connected to source <code>id</code>. This needs firmware version 61 or higher.
	 * Not approved for serial connections.
	 * <br>For details see <b>Multi_SetNSZCalibrationData</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()}
	 * @param calibrationValues calibration value per channel (max. 24) in nanoseconds
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> successful operation
	 * <br><b>KK_Err:</b> command failed, KK_Result.data contains error message
	 * <br><b>KK_ParamError:</b> invalid parameter, KK_Result.data contains error message.
	 * <br><b>KK_NotSupported:</b> K+K device does not support this function,
	 *  KK_Result.data contains error message. 
	 *   
	 * @since K+K library version 18.1.10
	 */
	public KK_Result setNSZCalibrationData(int id, double[] calibrationValues) {
		KK_Result result = new KK_Result();
		result.resultCode = ErrorCode.KK_No_Err;

		// check calibrationValues
		if (calibrationValues.length > 24) {
			result.resultCode = ErrorCode.KK_ParamError;
			result.data = "Calibration values limited to 24 channels";
		}
		else if (calibrationValues.length == 0) {
			result.resultCode = ErrorCode.KK_ParamError;
			result.data = "Calibration values missing";
		}
		if (result.resultCode != ErrorCode.KK_No_Err)
			return result;
		
		// convert calibrationValues to String
		StringBuilder sb = new StringBuilder();
		Formatter formatter = new Formatter(sb, kkLocale);
		formatter.format("%.3f", calibrationValues[0]);
		for (int i = 1; i < calibrationValues.length; i++)
			formatter.format("; %.3f", calibrationValues[i]);
		String s = sb.toString();
		formatter.close();
		
		// set calibration
		int res = KK_Lib.INSTANCE.Multi_SetNSZCalibrationData(id, s);
		if (res == CKK_DLL_NoError)
			result.resultCode = ErrorCode.KK_No_Err;
		else {
			if (res == CKK_DLL_SourceNotFound) {
				result.resultCode = ErrorCode.KK_ParamError;
				result.data = "Source-ID "+id+" not found";
			}
			else if (res == CKK_DLL_NotSupported) {
				result.resultCode = ErrorCode.KK_NotSupported;
				// serial connection?
				if (KK_Lib.INSTANCE.Multi_IsSerialDevice(id))
					result.data = "Serial connection";
				else result.data = "Firmware version < 62";
			}
			else {
				result.resultCode = ErrorCode.KK_ParamError;
				result.data = "Conversion error, set decimal separator!";
			}
		}
		return result;
	}
	
	/*
	 ***************************************************************************
	 * FHR Settings - helper classes
	 ***************************************************************************
	 */

	/**
	 * Defines FHR setting for one channel
	 * @author loryn
	 *
	 */
	public class FHR_Data {
	
		public String nominalFreq;
		public String loFreq;
		public boolean enabled;
		
		/**
		 * Creates instance of class FHR_Data with all members cleared 
		 */
		public FHR_Data() {
			super();
			clear();
		}
		
		/**
		 * Creates instance of class FHR_Data and fills members with string representation s
		 * Throws IllegalArgumentException, if s is not a valid string representation
		 * @param s string representation of FHR_Data or empty string
		 */
		public FHR_Data(String s) throws IllegalArgumentException {
			super();
			if (!fromString(s)) {
				throw new IllegalArgumentException("Not a valid FHR_Data string representation: "+s);
			}
		}
		
		/**
		 * Clears all members
		 */
		public void clear() {
			nominalFreq = "";
			loFreq 		= "";
			enabled 	= false;
		}
		
		/**
		 * Returns string representation of FHR_Data: members separated by ';'
		 */
		public String toString() {
			String s;
			if (enabled)
				s = "1";
			else
				s = "0";
			
			return nominalFreq+';'+loFreq+';'+s;
		}
		
		/**
		 * Converts string representation to FHR_Data
		 * @param s string representation of FHR_Data or empty string
		 * @return true, if s is a valid string representation
		 */
		public boolean fromString(String s) {
			clear();
			if ((s == null) || (s == ""))
				return true;
			
			String[] l = s.split(";");
			if (l.length < 3)
				return false;
			
			nominalFreq = l[0];
			loFreq = l[1];
			if (l[2].equals("0"))
				enabled = false;
			else
				enabled = true;
			return true;
		}
	}
	
	/**
	 * Defines FHR settings for 24 channels
	 */
	public class FHR_Settings {
		
		/**
		 * FHR setting per channel
		 */
		public FHR_Data[] fhrChannels;
		
		/**
		 * Creates instance of FHR_Settings with all channels cleared
		 */
		public FHR_Settings() {
			super();
			fhrChannels = new FHR_Data[24];
			for (int i = 0; i < fhrChannels.length; i++)
				fhrChannels[i] = new FHR_Data();
		}
		
		/**
		 * Creates instance of FHR_Settings and fills channels with string representation s
		 * Throws IllegalArgumentException, if s is not a valid string representation
		 * @param s string representation of FHR_Settings or empty string
		 */
		public FHR_Settings(String s) throws IllegalArgumentException {
			super();
			if (! fromString(s))
				throw new IllegalArgumentException("Not a valid FHR_Settings string representation: "+s);
		}
		
		/**
		 * Clears all channels
		 */
		public void clear() {
			for (int i = 0; i < fhrChannels.length; i++)
				fhrChannels[i].clear();
		}
		
		/**
		 * Returns string representation of FHR_Settings: string representations per channel separated by '/'
		 */
		public String toString() {
			StringBuilder sb = new StringBuilder(fhrChannels[0].toString());
			for (int i = 1; i < fhrChannels.length; i++)
				sb.append('/').append(fhrChannels[i].toString());
			return sb.toString();
		}
		
		/**
		 * Converts string representation to FHR_Settings, limits s to 24 channels
		 * @param s string representation of FHR_Settings or empty string
		 * @return true, if s is a valid string representation
		 */
		public boolean fromString(String s) {
			clear();
			boolean aBool = true;
			
			if ((s == null) || (s == ""))
				return aBool;
			
			String[] l = s.split("/");
			for (int i = 0; (i < l.length) && (i < 24); i++) {
				if (!fhrChannels[i].fromString(l[i]))
					aBool = false;
				
			}
			return aBool;
		}
	}

	public FHR_Settings fhrData;
	
	/*
	 ***************************************************************************
	 * FHR Settings
	 ***************************************************************************
	 */
	
	/**
	 * Requests FHR settings from K+K device connected to source <code>id</code>.
	 * K+K device responds with 0x7902-Report including requested data.
     * Needs firmware version 67 or higher.
     * For details see Multi_ReadFHRData in K+K library manual.
     * 
	 * @param id source identifier returned by {@link #createMultiSource()}
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> successful operation
	 * <br><b>KK_Err:</b> command failed, KK_Result.data contains error message
	 * <br><b>KK_ParamError:</b> invalid parameter, KK_Result.data contains error message.
	 * <br><b>KK_NotSupported:</b> K+K device does not support this function,
	 *  KK_Result.data contains error message. 
	 *   
	 * @since K+K library version 19.1.2
	 */
	public KK_Result getFHRSettings(int id) {
		KK_Result result = new KK_Result();
		result.resultCode = ErrorCode.KK_No_Err;
		
		// request FHR settings
		int res = KK_Lib.INSTANCE.Multi_ReadFHRData(id);
		if (res == CKK_DLL_NoError)
			result.resultCode = ErrorCode.KK_No_Err;
		else {
			if (res == CKK_DLL_SourceNotFound) {
				result.resultCode = ErrorCode.KK_ParamError;
				result.data = "Source-ID "+id+" not found";
			}
			else if (res == CKK_DLL_NotSupported) {
				result.resultCode = ErrorCode.KK_NotSupported;
				result.data = "Firmware version < 67";
			}
			else {
				result.resultCode = ErrorCode.KK_Err;
				result.data = "Command failed";
			}
		}
		return result;
	}

	/**
	 * Sets FHR settings <code>fhrSettings</code> in K+K device 
	 * connected to source <code>id</code>. This needs firmware version 67 or higher.
	 * Not approved for serial connections.
	 * <br>For details see <b>Multi_SetFHRnData</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()}
	 * @param fhrSettings described in helper class {@link FHR_Settings}
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> successful operation
	 * <br><b>KK_Err:</b> command failed, KK_Result.data contains error message
	 * <br><b>KK_ParamError:</b> invalid parameter, KK_Result.data contains error message.
	 * <br><b>KK_NotSupported:</b> K+K device does not support this function,
	 *  KK_Result.data contains error message. 
	 *   
	 * @since K+K library version 19.1.2
	 */
	public KK_Result setFHRSettings(int id, FHR_Settings fhrSettings) {
		KK_Result result = new KK_Result();
		result.resultCode = ErrorCode.KK_No_Err;

		// check fhrSettings
		if (fhrSettings == null) {
			result.resultCode = ErrorCode.KK_ParamError;
			result.data = "FHR settings missing";
			return result;
		}
		
		// convert data to String
		String s = fhrSettings.toString();
		
		// set FHR settings
		int res = KK_Lib.INSTANCE.Multi_SetFHRData(id, s);
		if (res == CKK_DLL_NoError)
			result.resultCode = ErrorCode.KK_No_Err;
		else {
			if (res == CKK_DLL_SourceNotFound) {
				result.resultCode = ErrorCode.KK_ParamError;
				result.data = "Source-ID "+id+" not found";
			}
			else if (res == CKK_DLL_NotSupported) {
				result.resultCode = ErrorCode.KK_NotSupported;
				// serial connection?
				if (KK_Lib.INSTANCE.Multi_IsSerialDevice(id))
					result.data = "Serial connection";
				else result.data = "Firmware version < 67";
			}
			else {
				result.resultCode = ErrorCode.KK_ParamError;
				result.data = "Format error";
			}
		}

		return result;
	}
	

	/*
	 ***************************************************************************
	 * open, close
	 ***************************************************************************
	 */

	/**
	 * Opens connection for source <code>id</code> to K+K device or K+K server 
	 * or read data from file according 
	 * to <code>connectionString</code>.
	 * A previous opened connection for source <code>id</code> will be closed.
	 * <br>For details see <b>Multi_OpenConnection</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * @param connectionString 
	 * 	String describes connection to open:
	 *
	 * <p><b>A) Local device:</b>
	 * <br>A.1 <b>serial port</b>
	 * <br><i>serialPort</i> with
	 * <br><i>serialPort</i>: serial port name delivered by {@link #enumerateDevices(byte)}
	 * <br>e.g.: (Windows) 'COM1', (Linux) '/dev/ttyS0'
	 * <br><b>Linux:</b> Caller must belong to group <i>dialout</i>
	 * <br>A.2: <b>USB</b>
	 * <br><i>deviceName</i> with
	 * <br><i>deviceName</i>: USB device name delivered by {@link #enumerateDevices(byte)}
	 * <br><b>Linux:</b> Caller must belong to group <i>root</i>
	 * <br>A.3 <b>CAN</b>:
	 * <br>CAN:MENLO<i>Node</i> with
	 * <br><i>Node</i>: CAN device node number in HEX (0..F), part of CAN-ID, hard coded setting
	 * for Menlo-CAN via PCAN-Basic API
	 * <br>Available only under Windows with PCAN-USB adaptor
	 * 
	 * <p><b>B) Connection to K+K device via network:</b>
	 * <br><i>IPaddress</i>:<i>port</i> with
	 * <br><i>IPaddress</i>: IPv4 address of K+K device, e.g. 192.168.178.98
	 * <br><i>port</i>: port number (16 bit value) of own TCP server socket to receive measurement data
	 * <br><i>:port</i> may be missing, system assigns port number
	 * 
	 * <p><b>C) Connection to TCP server</b>:
	 * <br>TCP:<i>IPaddress</i>:<i>port</i> with
	 * <br><i>IPaddress</i>: IPv4 address of K+K TCP server, e.g. 192.168.178.98
	 * <br><i>port</i>: port number (16 bit value) of K+K TCP server
	 * <br>The TCP server must have been started by another K+K application
	 * <br>Connects to TCP server and receives reports on DLL level
	 * 
	 * <p><b>D) File:</b>
	 * <br><i>filename</i>[ LOOP_FOR_EVER] with
	 * <br><i>filename</i>: filename of binary file or text file with measurement data
	 * <br>[ LOOP_FOR_EVER] optional: read data from file in an endless loop, starting again from begin
	 * <br>For test purposes only. Does not connect to device, but reads measurement data from file.
	 * 
	 * <p><b>User id:</b>
	 * <br>K+K devices can manage up to 4 connections simultaneously. Every connection has its own
	 * user id. This user id can be selected by adding
	 * <br>USERID=<i>nr</i> with
	 * <br><i>nr</i>: specifies user id = 1..4
	 * <br>USERID takes effect only by connections to K+K device, <b>but not on serial connections</b>.
	 * 
	 * @param blockingIO open connection with blocking read and write operations
	 * @param contextMode open connection in tools mode: no measurement data is send
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> connection successful opened
	 * <br><b>KK_ParamError:</b> invalid id, KK_Result.data contains error message.
	 * <br><b>KK_Err:</b> open failed, KK_Result.data contains error message.
	 *   
	 * @since K+K library version 18.0
	 */
	public KK_Result openConnection(int id, String connectionString, boolean blockingIO, boolean contextMode) {
		KK_Result result = new KK_Result();
		result.resultCode = ErrorCode.KK_No_Err;
		
		// convert connectionString to ASCII byte array
		byte[] connection = connectionString.getBytes(StandardCharsets.US_ASCII);
		System.arraycopy(connection, 0, buffer, 0, connection.length);
		// append 0-byte
		buffer[connection.length] = 0;
		// open port
		int res = KK_Lib.INSTANCE.Multi_OpenConnectionEx(id, buffer, blockingIO, contextMode, null);
		if (res != CKK_DLL_NoError) {
			// buffer contains error message -> convert to string
			result.data = bufferToString(buffer);
			if (res == CKK_DLL_SourceNotFound)
				result.resultCode = ErrorCode.KK_ParamError;
			else 
				result.resultCode = ErrorCode.KK_Err;
		}

		return result;
	}
	
	/**
	 * Convenience method, same as {@link #openConnection(int, String, boolean, boolean)} with blockingIO
	 * and contextMode = false
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * @param connectionString String describes connection to open
	 *
	 * @return see {@link #openConnection(int, String, boolean, boolean)}
	 * @since K+K library version 18.0
	 */
	public KK_Result openConnection(int id, String connectionString) {
		return openConnection(id, connectionString, false, false);
	}
	
	/**
	 * Closes a previously opened connection ({@link #openConnection(int, String, boolean, boolean)})
	 * for source <code>id</code>.
	 * <br>For details see <b>Multi_CloseConnection</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()}
	 *  
	 * @since K+K library version 18.0
	 */
	public void closeConnection(int id) {
		KK_Lib.INSTANCE.Multi_CloseConnection(id);
	}

	/*
	 ***************************************************************************
	 * read, write
	 ***************************************************************************
	 */
	
	private Locale kkLocale = Locale.getDefault();
	
	/**
	 * Library K+K library reports received measurement data (floating point values) as
	 * string. Use this function to set decimal separator used to convert floating point
	 * numbers into string for source <code>id</code>.
	 * <br>For details see <b>Multi_SetDecimalSeparator</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * @param decimalSeparator new value for decimal separator, must be "." or
	 * ",", other values are ignored
	 * 
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> successful operation 
	 * <br><b>KK_ParamError:</b> invalid id
	 * 
	 * @since K+K library version 18.0
	 */
	public KK_Result setDecimalSeparator(int id, String decimalSeparator) {
		KK_Result result = new KK_Result();
		result.resultCode = ErrorCode.KK_No_Err;
		byte[] separatorBytes = decimalSeparator.getBytes(StandardCharsets.US_ASCII);
		
		int res = KK_Lib.INSTANCE.Multi_SetDecimalSeparator(id, separatorBytes[0]);
		if (res == CKK_DLL_SourceNotFound)
			result.resultCode = ErrorCode.KK_ParamError;
		else if (decimalSeparator == ".")
			kkLocale = Locale.ENGLISH;
		else
			kkLocale = Locale.GERMAN;
		
		return result;
	}

	/**
	 * Depending on used measurement card, K+K device reports NSZ measurements as single NSZ
	 * or double NSZ. Use this function call to set single or double NSZ for source <code>id</code>.
	 * <br>For details see <b>Multi_SetNSZ</b> in K+K library manual.
	 *  
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * @param nsz 1 = single NSZ, 2 = double NSZ, other values are ignored
	 * 
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> successful operation 
	 * <br><b>KK_ParamError:</b> invalid id
	 * 
	 * @since K+K library version 18.0
	 */
	public KK_Result setNSZ(int id, int nsz) {
		KK_Result result = new KK_Result();
		result.resultCode = ErrorCode.KK_No_Err;
		
		int res = KK_Lib.INSTANCE.Multi_SetNSZ(id, nsz);
		if (res == CKK_DLL_SourceNotFound)
			result.resultCode = ErrorCode.KK_ParamError;
		
		return result;
	}

	/**
	 * Gets next report for source <code>id</code> received via current connection from 
	 * K+K device or K+K server and write it to KK_Result.data.
	 * <br>For details see <b>Multi_GetReport</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()}
	 *  
	 * @return KK_Result.data contains next report or <code>null</code> if no reports available or
	 * error message if KK_Result.resultCode != KK_No_Err
	 * <br>Values for KK_Result.resultCode:
	 * <br><b>KK_No_Err:</b> Reading successful
	 * <br><b>KK_Err_BufferTooSmall:</b> Reading successful, but String in KK_Result.data is 
	 * truncated to 1024 characters.
	 * <br><b>KK_Err:</b> An error occurred, see error message in KK_Result.data
	 * <br><b>KK_Err_Write:</b> Writing to device/server failed. Connection is closed. 
	 * <br><b>KK_Err_ServerDown:</b> Server is not online. Connection is closed.
	 * <br><b>KK_Err_DeviceNotConnected:</b> No connection to K+K device (connection has not been
	 * established or usb cable is unplugged)
	 * <br><b>KK_HardwareFault:</b> Fault in measurement hardware. Connection to K+K device is closed.
	 * <br><b>KK_ParamError:</b> invalid id
	 * <br><b>KK_Reconnected:</b> Reconnection of an interrupted connection with data loss
	 * 
	 * @since K+K library version 18.0
	 */
	public KK_Result getReport(int id) {
		KK_Result result = new KK_Result();

		int res = KK_Lib.INSTANCE.Multi_GetReport(id, buffer);
		result.data = bufferToString(buffer);
		// check returnCode
		switch (res) {
		case CKK_DLL_NoError:
			result.resultCode = ErrorCode.KK_No_Err;
			break;
		case CKK_DLL_BufferTooSmall:
			result.resultCode = ErrorCode.KK_Err_BufferTooSmall;
			break;
		case CKK_DLL_Error:
			result.resultCode = ErrorCode.KK_Err;
			break;
		case CKK_DLL_WriteError:
			result.resultCode = ErrorCode.KK_Err_Write;
			break;
		case CKK_DLL_ServerDownError:
			result.resultCode = ErrorCode.KK_Err_ServerDown;
			break;
		case CKK_DLL_DeviceNotConnected:
			result.resultCode = ErrorCode.KK_Err_DeviceNotConnected;
			break;
		case CKK_DLL_HardwareFault:
			result.resultCode = ErrorCode.KK_HardwareFault;
			break;
		case CKK_DLL_SourceNotFound:
			result.resultCode = ErrorCode.KK_ParamError;
			break;
		case CKK_DLL_Reconnected:
			result.resultCode = ErrorCode.KK_Reconnected;
			break;
		default:
			// maybe new error numbers in future
			result.resultCode = ErrorCode.KK_Err;
		}

		
		return result;
	}
	
	/**
	 * The 100ms timestamps received from the K+K device are passed on to the application
	 * by default with Report 7000.
	 * Here it is now possible to switch to Report 7016.
	 * <br>For details see <b>Multi_SetSend7016</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()}
	 * @param value
	 * <br>True: send Report 7016
	 * <br>False: send Report 7000 (Default)
	 * 
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> successful operation 
	 * <br><b>KK_ParamError:</b> invalid id
	 * 
	 * @since K+K library version 19.3.1
	 */
	public KK_Result setSend7016(int id, boolean value) {
		KK_Result result = new KK_Result();
		
		int res = KK_Lib.INSTANCE.Multi_SetSend7016(id, value);
		// check returnCode
		switch (res) {
		case CKK_DLL_NoError:
			result.resultCode = ErrorCode.KK_No_Err;
			break;
		case CKK_DLL_SourceNotFound:
			result.resultCode = ErrorCode.KK_ParamError;
			break;
		default:
			// maybe new error numbers in future
			result.resultCode = ErrorCode.KK_Err;
		}
		return result;
	}
	
	/**
	 * Get response to previous READ command (85 40) for source <code>id</code> via current 
	 * connection from K+K device and write it to KK_Result.data.
	 * <br>For details see <b>Multi_GetToolsData</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()}
	 *  
	 * @return KK_Result.data contains read response or <code>null</code> if no response available
	 * or error message if KK_Result.resultCode != KK_No_Err
	 * <br>Values for KK_Result.resultCode:
	 * <br><b>KK_No_Err:</b> Reading successful
	 * <br><b>KK_Err_BufferTooSmall:</b> Reading successful, but String in KK_Result.data is 
	 * truncated to 1024 characters.
	 * <br><b>KK_Err:</b> An error occurred, see error message in KK_Result.data
	 * <br><b>KK_Err_DeviceNotConnected:</b> No connection to K+K device (connection has not been
	 * established or usb cable is unplugged)
	 * <br><b>KK_ParamError:</b> invalid id
	 * 
	 * @since K+K library version 18.0
	 */
	public KK_Result getToolsData(int id) {
		KK_Result result = new KK_Result();

		int res = KK_Lib.INSTANCE.Multi_GetToolsData(id, buffer);
		result.data = bufferToString(buffer);
		// check returnCode
		switch (res) {
		case CKK_DLL_NoError:
			result.resultCode = ErrorCode.KK_No_Err;
			break;
		case CKK_DLL_BufferTooSmall:
			result.resultCode = ErrorCode.KK_Err_BufferTooSmall;
			break;
		case CKK_DLL_Error:
			result.resultCode = ErrorCode.KK_Err;
			break;
		case CKK_DLL_DeviceNotConnected:
			result.resultCode = ErrorCode.KK_Err_DeviceNotConnected;
			break;
		case CKK_DLL_SourceNotFound:
			result.resultCode = ErrorCode.KK_ParamError;
			break;
		default:
			// maybe new error numbers in future
			result.resultCode = ErrorCode.KK_Err;
		}
		
		return result;
		
	}
	
	/**
	 * Gets count of buffered (not yet send) commands for source <code>id</code>.
	 * <br>For details see <b>Multi_GetPendingCmdsCount</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * 
	 * @return 
	 * 0: no waiting commands or error (invalid id, no current connection)
	 * <br>0: count of waiting commands for source <code>id</code>
	 *   
	 * @since K+K library version 18.0
	 */
	public int getPendingCmdsCount(int id) {
		return KK_Lib.INSTANCE.Multi_GetPendingCmdsCount(id);
	}
	
	/**
	 * Sets limit of command waiting queue for source <code>id</code> to
	 * <code>limit</code>. For an unlimited waiting queue set limit to 0.
	 * Default = unlimited.
	 * <br>For details see <b>Multi_SetCommandLimit</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * @param limit length of waiting queue, 0=unlimited, default=unlimited
	 * 
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> successful operation 
	 * <br><b>KK_ParamError:</b> invalid id
	 *   
	 * @since K+K library version 18.0
	 */
	public KK_Result setCommandLimit(int id, int limit) {
		KK_Result result = new KK_Result();	
		result.resultCode = ErrorCode.KK_No_Err;

		int res =  KK_Lib.INSTANCE.Multi_SetCommandLimit(id, limit);
		if (res == CKK_DLL_SourceNotFound)
			result.resultCode = ErrorCode.KK_ParamError;
		
		return result;
	}
	
	/**
	 * Adds <code>command</code> to waiting queue of source <code>id</code>.
	 * If there is no current connection or waiting queue is full, command is rejected
	 * (KK_CmdIgnored). 
	 * <br>Usually command is send with next {@link #getReport(int)} call and
	 * in case of an error an error message (KK_Err_Write) is replied there. 
	 * <br>With network connection, command is send immediately (no waiting queue), in case
	 * of an error an error message is returned here.
	 * <br><b>Hint:</b> Command is <b>not send</b> if connection is file or connection to TCP server. 
	 * <br>For details see <b>Multi_SendCommand</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * @param command command bytes to send
	 * 
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> operation successful
	 * <br>all other codes are error numbers and KK_Result.data contains error message:
	 * <br><b>KK_ParamError:</b> invalid parameter value
	 * <br><b>KK_CmdIgnored:</b> command rejected   
	 * <br><b>KK_Err:</b> an error occurred
	 * 
	 * @since K+K library version 18.0
	 */
	public KK_Result sendCommand(int id, byte[] command) {
		KK_Result result = new KK_Result();	
		if (command == null || command.length == 0) {
			result.resultCode = ErrorCode.KK_ParamError;
			result.data = "Invalid parameter command";
			return result;
		}
		
		// copy into buffer
		int len = command.length;
		System.arraycopy(command, 0, buffer, 0, len);
		// append 0-byte
		buffer[len] = 0;
		
		int res = KK_Lib.INSTANCE.Multi_SendCommand(id, buffer, len);
		if (res == CKK_DLL_NoError)
			result.resultCode = ErrorCode.KK_No_Err;
		else {
			// buffer contains error message -> convert to string
			result.data = bufferToString(buffer);
			if (res == CKK_DLL_SourceNotFound)
				result.resultCode = ErrorCode.KK_ParamError;
			else if (res == CKK_DLL_CmdIgnored)
				result.resultCode = ErrorCode.KK_CmdIgnored;
			else
				result.resultCode = ErrorCode.KK_Err;
		}
		
		return result;
	}
	
	/**
	 * Remote login procedure on K+K device connected to source <code>id</code>.
	 * Remote login is needed to change flash memory of K+K device with
	 * network connections. If source connection is not via network, remote
	 * login is ignored without error.
	 * <br>For details see <b>Multi_RemoteLogin</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * @param password remote password
	 * 
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> operation successful
	 * <br>all other codes are error numbers and KK_Result.data contains error message:
	 * <br><b>KK_ParamError:</b> invalid parameter value
	 * <br><b>KK_CmdIgnored:</b> command rejected   
	 * <br><b>KK_Err:</b> invalid password
	 * 
	 * @since K+K library version 18.1.10
	 */
	public KK_Result remoteLogin(int id, int password) {
		KK_Result result = new KK_Result();	
		int res = KK_Lib.INSTANCE.Multi_RemoteLogin(id, password, buffer);
		if (res == CKK_DLL_NoError)
			result.resultCode = ErrorCode.KK_No_Err;
		else {
			// buffer contains error message -> convert to string
			result.data = bufferToString(buffer);
			if (res == CKK_DLL_SourceNotFound)
				result.resultCode = ErrorCode.KK_ParamError;
			else if (res == CKK_DLL_CmdIgnored)
				result.resultCode = ErrorCode.KK_CmdIgnored;
			else
				result.resultCode = ErrorCode.KK_Err;
		}

		return result;
	}
	
	/*
	 ***************************************************************************
	 * TCP server
	 ***************************************************************************
	 */

	/**
	 * Starts local TCP server for source <code>id</code> on port <code>aPort</code>. 
	 * If aPort = 0, port number is assigned by system and returned in KK_Result.intValue
	 * <br>For details see <b>Multi_StartTcpServer</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * @param aPort port number for TCP server
	 * 
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> TCP server is running
	 * <br><b>KK_ParamError:</b> invalid id, KK_Result.data contains error message.
	 * <br><b>KK_Err:</b> an error occurred, KK_Result.data contains error message.
	 * 
	 * @since K+K library version 18.0
	 */
	public KK_Result startTcpServer(int id, short aPort) {
		KK_Result result = new KK_Result();
		short[] port = new short[1];
		port[0] = aPort;
		int res = KK_Lib.INSTANCE.Multi_StartTcpServer(id, port);
		if (res == CKK_DLL_NoError) {
			result.resultCode = ErrorCode.KK_No_Err;
			result.intValue = port[0];
		} 
		else if (res == CKK_DLL_SourceNotFound) {
			result.resultCode = ErrorCode.KK_ParamError;
			// set error message
			result.data = ERR_SOURCE_NOT_FOUND+id+" not found";
		}
		else {
			result.resultCode = ErrorCode.KK_Err;
			// set error message
			result.data = KK_Lib.INSTANCE.Multi_GetTcpServerError(id);
		}
		
		return result;
	}
	
	/**
	 * Stops local TCP server for source <code>id</code>.
	 * All client connections are disconnected.
	 * <br>For details see <b>Multi_SopTcpServer</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * 
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> successful operation
	 * <br><b>KK_ParamError:</b> invalid id, KK_Result.data contains error message.
	 * <br><b>KK_Err:</b> an error occurred, KK_Result.data contains error message.
	 * 
	 * @since K+K library version 18.0
	 */
	public KK_Result stopTcpServer(int id) {
		KK_Result result = new KK_Result();
		
		int res = KK_Lib.INSTANCE.Multi_StopTcpServer(id);
		if (res == CKK_DLL_NoError) {
			result.resultCode = ErrorCode.KK_No_Err;
		} 
		else if (res == CKK_DLL_SourceNotFound) {
			result.resultCode = ErrorCode.KK_ParamError;
			// set error message
			result.data = ERR_SOURCE_NOT_FOUND+id+" not found";
		}
		else {
			result.resultCode = ErrorCode.KK_Err;
			// set error message
			result.data = KK_Lib.INSTANCE.Multi_GetTcpServerError(id);
		}
		
		return result;
	}
	
	/**
	 * Defines log entry type values for method {@link KK_Library#tcpReportLog(int, String, TcpLogType)}
	 * @author Loryn
	 *
	 */
	public enum TcpLogType {
		/**
		 * Log entry type phase data, corresponds with reportmode {@link KK_Library#LOG_MODE_PHASE}
		 */
		PhaseData,
		/**
		 * Log entry type frequnce data, corresponds with reportmode {@link KK_Library#LOG_MODE_FREQ}
		 */
		FreqData,
		/**
		 * Log entry type phase difference data, corresponds with reportmode
		 * {@link KK_Library#LOG_MODE_PHASEDIFF}
		 */
		PhaseDiffData,
		/**
		 * Log entry type NSZ data, corresponds with reportmode
		 * {@link KK_Library#LOG_MODE_NSZ}
		 * @since 18.01.10
		 */
		NSZData,
		/**
		 * Log entry type NSZ difference data, corresponds with reportmode
		 * {@link KK_Library#LOG_MODE_NSZDIFF}
		 * @since 18.01.10
		 */
		NSZDiffData,
		/**
		 * Log entry type phase difference data to predecessor, corresponds with reportmode
		 * {@link KK_Library#LOG_MODE_PHASEPREDECESSOR}
		 * @since 18.02.04
		 */
		PhasePredecessorData,
		/**
		 * Log entry type user defined data 1, corresponds with reportmode
		 * {@link KK_Library#LOG_MODE_USER1}
		 * @since 18.02.04
		 */
		UserData1,
		/**
		 * Log entry type user defined data 2, corresponds with reportmode
		 * {@link KK_Library#LOG_MODE_USER2}
		 * @since 18.02.04
		 */
		UserData2
	}
	
	/**
	 * Transmits log entry <code>data</code> for LOG level type <code>logType</code> to TCP server
	 * on source <code>id</code>.
	 * TCP server proceeds data to connected TCP receiver clients.
	 * <br>If an error occurs (id invalid, source without tcp server, data=null), call is ignored.
	 * <br>For details see <b>Multi_TcpReportLog</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * @param data Log entry to transmit to connected clients
	 * @param logType specifies log entry type see {@link TcpLogType}
	 * 
	 * @since K+K library version 18.0
	 */
	public void tcpReportLog(int id, String data, TcpLogType logType) {
		int type;
		switch (logType) {
		case PhaseData: 	type = 0; break;
		case FreqData: 		type = 1; break;
		case PhaseDiffData: type = 2; break;
		case NSZData:		type = 3; break;
		case NSZDiffData:	type = 4; break;
		case PhasePredecessorData:
							type = 5; break;
		case UserData1:		type = 6; break;
		case UserData2:		type = 7; break;
		default: 			type = 0; break;
		}
		KK_Lib.INSTANCE.Multi_TcpReportLog(id, data, type);
	}
	
	/*
	 ***************************************************************************
	 * connection to TCP server
	 ***************************************************************************
	 */

	/**
	 * Report mode phase log data (log entries given to TCP server with {@link TcpLogType#PhaseData})
	 */
	public static final String LOG_MODE_PHASE		= "PHASELOG";
	/**
	 * Report mode frequency log data (log entries given to TCP server with {@link TcpLogType#FreqData})
	 */
	public static final String LOG_MODE_FREQ		= "FREQLOG";
	/**
	 * Report mode phase difference log data (log entries given to TCP server with
	 * {@link TcpLogType#PhaseDiffData})
	 */
	public static final String LOG_MODE_PHASEDIFF	= "PHASEDIFFLOG";
	/**
	 * Report mode NSZ log data (log entries given to TCP server with {@link TcpLogType#NSZData})
	 * @since 18.01.10
	 */
	public static final String LOG_MODE_NSZ  		= "NSZLOG";
	/**
	 * Report mode NSZ difference log data (log entries given to TCP server with {@link TcpLogType#NSZDiffData})
	 * @since 18.01.10
	 */
	public static final String LOG_MODE_NSZDIFF		= "NSZDIFFLOG";
	/**
	 * Report mode phase difference to predecessor log data (log entries given to TCP server with {@link TcpLogType#PhasePredecessorData})
	 * @since 18.02.04
	 */
	public static final String LOG_MODE_PHASEPREDECESSOR  = "PHASEPREDECESSORLOG";
	/**
	 * Report mode user defined log data 1 (log entries given to TCP server with {@link TcpLogType#UserData1})
	 * @since 18.02.04
	 */
	public static final String LOG_MODE_USER1		= "USER1LOG";
	/**
	 * Report mode user defined log data 2 (log entries given to TCP server with {@link TcpLogType#UserData2})
	 * @since 18.02.04
	 */
	public static final String LOG_MODE_USER2		= "USER2LOG";

	/**
	 * Opens connection for source <code>id</code> to K+K TCP server on address <code>ipPort</code>
	 * for log level entries.
	 * <br>For details see <b>Multi_OpenTcpLog</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * @param ipPort <i>IPaddress</i>:<i>port</i> with
	 * <br><i>IPaddress</i>: IPv4 address of K+K TCP server, e.g. 192.168.178.98 or 127.0.0.1 
	 * for local host
	 * <br><i>port</i>: port number (16 bit value) of TCP server.
	 * @param mode report mode to receive: {@link KK_Library#LOG_MODE_PHASE}, {@link KK_Library#LOG_MODE_FREQ},
	 * {@link KK_Library#LOG_MODE_PHASEDIFF}, {@link KK_Library#LOG_MODE_NSZ}, {@link KK_Library#LOG_MODE_NSZDIFF},
	 * {@link KK_Library#LOG_MODE_PHASEPREDECESSOR}, {@link KK_Library#LOG_MODE_USER1}, {@link KK_Library#LOG_MODE_USER2}
	 * 
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> connection successful opened
	 * <br><b>KK_ParamError:</b> invalid id, KK_Result.data contains error message.
	 * <br><b>KK_Err:</b> open failed, KK_Result.data contains error message.
	 * 
	 * @since K+K library version 18.0
	 */
	public KK_Result openTcpLog(int id, String ipPort, String mode) {
		KK_Result result = new KK_Result();

		// convert ipPort to ASCII byte array
		byte[] connection = ipPort.getBytes(StandardCharsets.US_ASCII);
		System.arraycopy(connection, 0, buffer, 0, connection.length);
		// append 0-byte
		buffer[connection.length] = 0;
		// open log
		int res = KK_Lib.INSTANCE.Multi_OpenTcpLog(id, buffer, mode);
		// check error
		if (res == CKK_DLL_NoError)
			result.resultCode = ErrorCode.KK_No_Err;
		else {
			// buffer contains error message
			result.data = bufferToString(buffer);
			if (res == CKK_DLL_SourceNotFound)
				result.resultCode = ErrorCode.KK_ParamError;
			else
				result.resultCode = ErrorCode.KK_Err;
		}

		return result;
	}

	/**
	 * Opens connection for source <code>id</code> to K+K TCP server on address <code>ipPort</code>
	 * for log level entries with prepended UTC timestamp.
	 * <br>For details see <b>Multi_OpenTcpLogTime</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * @param ipPort <i>IPaddress</i>:<i>port</i> with
	 * <br><i>IPaddress</i>: IPv4 address of K+K TCP server, e.g. 192.168.178.98 or 127.0.0.1 
	 * for local host
	 * <br><i>port</i>: port number (16 bit value) of TCP server.
	 * @param mode report mode to receive: {@link KK_Library#LOG_MODE_PHASE}, {@link KK_Library#LOG_MODE_FREQ},
	 * {@link KK_Library#LOG_MODE_PHASEDIFF}, {@link KK_Library#LOG_MODE_NSZ}, {@link KK_Library#LOG_MODE_NSZDIFF},
	 * {@link KK_Library#LOG_MODE_PHASEPREDECESSOR}, {@link KK_Library#LOG_MODE_USER1}, {@link KK_Library#LOG_MODE_USER2}
	 * @param formatTime specifies format of UTC timestamp
	 * 
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> connection successful opened
	 * <br><b>KK_ParamError:</b> invalid id, KK_Result.data contains error message.
	 * <br><b>KK_Err:</b> open failed, KK_Result.data contains error message.
	 * 
	 * @since K+K library version 19.0.2
	 */
	public KK_Result openTcpLogTime(int id, String ipPort, String mode, String formatTime) {
		KK_Result result = new KK_Result();

		// convert ipPort to ASCII byte array
		byte[] connection = ipPort.getBytes(StandardCharsets.US_ASCII);
		System.arraycopy(connection, 0, buffer, 0, connection.length);
		// append 0-byte
		buffer[connection.length] = 0;
		// open log
		int res = KK_Lib.INSTANCE.Multi_OpenTcpLogTime(id, buffer, mode, formatTime);
		// check error
		if (res == CKK_DLL_NoError)
			result.resultCode = ErrorCode.KK_No_Err;
		else {
			// buffer contains error message
			result.data = bufferToString(buffer);
			if (res == CKK_DLL_SourceNotFound)
				result.resultCode = ErrorCode.KK_ParamError;
			else
				result.resultCode = ErrorCode.KK_Err;
		}

		return result;
	}

	/**
	 * Opens connection for source <code>id</code> to K+K TCP server on address <code>ipPort</code>
	 * for log level entries with or without prepended UTC timestamp, depending on <code>format</code>.
	 * <br>For details see <b>Multi_OpenTcpLogType</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * @param ipPort <i>IPaddress</i>:<i>port</i> with
	 * <br><i>IPaddress</i>: IPv4 address of K+K TCP server, e.g. 192.168.178.98 or 127.0.0.1 
	 * for local host
	 * <br><i>port</i>: port number (16 bit value) of TCP server.
	 * @param logType report mode to receive (0..7)
	 * @param formatTime specifies format of UTC timestamp, if not <code>null</code>
	 * 
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> connection successful opened
	 * <br><b>KK_ParamError:</b> invalid id, KK_Result.data contains error message.
	 * <br><b>KK_Err:</b> open failed, KK_Result.data contains error message.
	 * 
	 * @since K+K library version 19.2.0
	 */
	public KK_Result openTcpLogType(int id, String ipPort, int logType, String formatTime) {
		KK_Result result = new KK_Result();

		// convert ipPort to ASCII byte array
		byte[] connection = ipPort.getBytes(StandardCharsets.US_ASCII);
		System.arraycopy(connection, 0, buffer, 0, connection.length);
		// append 0-byte
		buffer[connection.length] = 0;
		// open log
		int res = KK_Lib.INSTANCE.Multi_OpenTcpLogType(id, buffer, logType, formatTime);
		// check error
		if (res == CKK_DLL_NoError)
			result.resultCode = ErrorCode.KK_No_Err;
		else {
			// buffer contains error message
			result.data = bufferToString(buffer);
			if (res == CKK_DLL_SourceNotFound)
				result.resultCode = ErrorCode.KK_ParamError;
			else
				result.resultCode = ErrorCode.KK_Err;
		}

		return result;
	}

	/**
	 * Closes previously opened connection ({@link #openTcpLog(int, String, String)}) 
	 * for source <code>id</code> to LOG level K+K TCP server.
	 * <br>If an error occurs (id invalid, no connection), call is ignored.
	 * <br>For details see <b>Multi_CloseTcpLog</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * 
	 * @since K+K library version 18.0
	 */
	public void closeTcpLog(int id) {
		KK_Lib.INSTANCE.Multi_CloseTcpLog(id);
	}

	/**
	 * Get next log level entry for source <code>id</code> received from K+K TCP server and
	 * write it to KK_Result.data.
	 * KK_Result.data = <code>null</code>, if no reports available.
	 * <br>For details see <b>Multi_GetTcpLog</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * 
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> operation successful
	 * <br><b>KK_Err_BufferTooSmall:</b> operation successful, but string in KK_Result.data is 
	 * truncated to 1024 characters.
	 * <br>all other codes are error numbers and KK_Result.data contains error message:
	 * <br><b>KK_ParamError:</b> invalid id
	 * <br><b>KK_Err_ServerDown:</b> Server is not online. Connection is closed
	 * <br><b>KK_Err_BufferOverflow:</b> Data was lost. App does not read fast enough.
	 * 
	 * @since K+K library version 18.0
	 */
	public KK_Result getTcpLog(int id) {
		KK_Result result = new KK_Result();
		
		// get log
		int res = KK_Lib.INSTANCE.Multi_GetTcpLog(id, buffer);
		// convert to string
		result.data = bufferToString(buffer);
		// check error
		if (res == CKK_DLL_NoError)
			result.resultCode = ErrorCode.KK_No_Err;
		else if (res == CKK_DLL_BufferTooSmall)
			result.resultCode = ErrorCode.KK_Err_BufferTooSmall;
		else if (res == CKK_DLL_SourceNotFound)
			result.resultCode = ErrorCode.KK_ParamError;
		else if (res == CKK_DLL_ServerDownError)
			result.resultCode = ErrorCode.KK_Err_ServerDown;
		else if (res == CKK_DLL_BufferOverflow)
			result.resultCode = ErrorCode.KK_Err_BufferOverflow;
		else // maybe new error numbers in future
			result.resultCode = ErrorCode.KK_Err;
		
		return result;
	}
	
	public KK_Result tcpAppData(int id, String data) {
		KK_Result result = new KK_Result();
		// convert data to ASCII byte array
		byte[] dataB = data.getBytes(StandardCharsets.US_ASCII);
		System.arraycopy(dataB, 0, buffer, 0, dataB.length);
		// append 0-byte
		buffer[dataB.length] = 0;
		// send data
		int res = KK_Lib.INSTANCE.Multi_TcpAppData(id, buffer);
		// convert to string
		result.data = bufferToString(buffer);
		// check error
		if (res == CKK_DLL_NoError)
			result.resultCode = ErrorCode.KK_No_Err;
		else if (res == CKK_DLL_SourceNotFound)
			result.resultCode = ErrorCode.KK_ParamError;
		else if (res == CKK_DLL_ServerDownError)
			result.resultCode = ErrorCode.KK_Err_ServerDown;
		else 
			result.resultCode = ErrorCode.KK_Err;			
		
		return result;
	}
	
	/*
	 ***************************************************************************
	 * generate test data
	 ***************************************************************************
	 */

	/**
	 * Start generating test data for source <code>id</code> in binary file.
	 * All received reports are written to binary file <i>Date_Time_dbgID</i>_KK_BinaryData.bin
	 * in set output directory {@link #setOutputPath(int, String)}.
	 * <br><b>Hint:</b> <code>DbgID</code> is used to get unique file names for
	 * distinct sources. This is needed in particular, if common output paths are used.
	 * <code>dbgOn</code> maybe null, if only one source is used.
	 * <br>For details see <b>Multi_StartSaveBinaryData</b> in K+K library manual.
	 *  
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * @param dbgID source identifier of binary file, part of file name
	 * 
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> operation successful
	 * <br><b>KK_ParamError:</b> invalid id
	 * <br><b>KK_Err_DeviceNotConnected:</b> no connection
	 * 
	 * @since K+K library version 18.0
	 */
	public KK_Result startSaveBinaryData(int id, String dbgID) {
		KK_Result result = new KK_Result();

		int res = KK_Lib.INSTANCE.Multi_StartSaveBinaryData(id, dbgID);
		// check error
		if (res == CKK_DLL_NoError)
			result.resultCode = ErrorCode.KK_No_Err;
		else if (res == CKK_DLL_SourceNotFound)
			result.resultCode = ErrorCode.KK_ParamError;
		else if (res == CKK_DLL_DeviceNotConnected)
			result.resultCode = ErrorCode.KK_Err_DeviceNotConnected;
		else // maybe new error numbers in future
			result.resultCode = ErrorCode.KK_Err;
		
		return result;
	}
	
	/**
	 * Stops generating binary test data for source <code>id</code>, closes binary file. 
	 * Binary file is closed too, when current connection is closed.
	 * <br>For details see <b>Multi_StopSaveBinaryData</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()}
	 *  
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> operation successful
	 * <br><b>KK_ParamError:</b> invalid id
	 * <br><b>KK_Err_DeviceNotConnected:</b> no connection
	 * 
	 * @since K+K library version 18.0
	 */
	public KK_Result stopSaveBinaryData(int id) {
		KK_Result result = new KK_Result();

		int res = KK_Lib.INSTANCE.Multi_StopSaveBinaryData(id);
		// check error
		if (res == CKK_DLL_NoError)
			result.resultCode = ErrorCode.KK_No_Err;
		else if (res == CKK_DLL_SourceNotFound)
			result.resultCode = ErrorCode.KK_ParamError;
		else if (res == CKK_DLL_DeviceNotConnected)
			result.resultCode = ErrorCode.KK_Err_DeviceNotConnected;
		else // maybe new error numbers in future
			result.resultCode = ErrorCode.KK_Err;
		
		return result;
	}

	/**
	 * Start generating test data for source <code>id</code> in text file.
	 * All received reports are written to text file <i>Date_Time_dbgID</i>_KK_ReportData.txt
	 * in set output directory {@link #setOutputPath(int, String)}.
	 * <br><b>Hint:</b> <code>DbgID</code> is used to get unique file names for
	 * distinct sources. This is needed in particular, if common output paths are used.
	 * <code>dbgOn</code> maybe null, if only one source is used.
	 * <br>For details see <b>Multi_StartSaveReportData</b> in K+K library manual.
	 *  
	 * @param id source identifier returned by {@link #createMultiSource()} 
	 * @param dbgID source identifier of text file, part of file name
	 * 
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> operation successful
	 * <br><b>KK_ParamError:</b> invalid id
	 * 
	 * @since K+K library version 18.0
	 */
	public KK_Result startSaveReportData(int id, String dbgID) {
		KK_Result result = new KK_Result();

		int res = KK_Lib.INSTANCE.Multi_StartSaveReportData(id, dbgID);
		// check error
		if (res == CKK_DLL_NoError)
			result.resultCode = ErrorCode.KK_No_Err;
		else if (res == CKK_DLL_SourceNotFound)
			result.resultCode = ErrorCode.KK_ParamError;
		else // maybe new error numbers in future
			result.resultCode = ErrorCode.KK_Err;
		
		return result;
	}

	/**
	 * Stops generating report test data for source <code>id</code>.
	 * <br>For details see <b>Multi_StopSaveReportData</b> in K+K library manual.
	 * 
	 * @param id source identifier returned by {@link #createMultiSource()}
	 *  
	 * @return see KK_Result.resultCode
	 * <br><b>KK_No_Err:</b> operation successful
	 * <br><b>KK_ParamError:</b> invalid id
	 * 
	 * @since K+K library version 18.0
	 */
	public KK_Result stopSaveReportData(int id) {
		KK_Result result = new KK_Result();

		int res = KK_Lib.INSTANCE.Multi_StopSaveReportData(id);
		// check error
		if (res == CKK_DLL_NoError)
			result.resultCode = ErrorCode.KK_No_Err;
		else if (res == CKK_DLL_SourceNotFound)
			result.resultCode = ErrorCode.KK_ParamError;
		else // maybe new error numbers in future
			result.resultCode = ErrorCode.KK_Err;
		
		return result;
	}

	/*
	 ***************************************************************************
	 * other
	 ***************************************************************************
	 */

}
